@@ -100,6 +100,7 @@ struct GMenuTreeDirectory
 
   guint only_unallocated : 1;
   guint is_root : 1;
+  guint is_nodisplay : 1;
 };
 
 typedef struct
@@ -117,6 +118,7 @@ struct GMenuTreeEntry
   char         *desktop_file_id;
 
   guint is_excluded : 1;
+  guint is_nodisplay : 1;
 };
 
 struct GMenuTreeSeparator
@@ -1009,6 +1011,14 @@ gmenu_tree_directory_get_tree (GMenuTreeDirectory *directory)
   return gmenu_tree_ref (root->tree);
 }
 
+gboolean
+gmenu_tree_directory_get_is_nodisplay (GMenuTreeDirectory *directory)
+{
+  g_return_val_if_fail (directory != NULL, FALSE);
+
+  return directory->is_nodisplay;
+}
+
 static void
 append_directory_path (GMenuTreeDirectory *directory,
 		       GString            *path)
@@ -1101,6 +1111,14 @@ gmenu_tree_entry_get_is_excluded (GMenuTreeEntry *entry)
   return entry->is_excluded;
 }
 
+gboolean
+gmenu_tree_entry_get_is_nodisplay (GMenuTreeEntry *entry)
+{
+  g_return_val_if_fail (entry != NULL, FALSE);
+
+  return entry->is_nodisplay;
+}
+
 GMenuTreeDirectory *
 gmenu_tree_header_get_directory (GMenuTreeHeader *header)
 {
@@ -1160,6 +1178,7 @@ gmenu_tree_directory_new (GMenuTreeDirectory *parent,
   retval->layout_info         = NULL;
   retval->contents            = NULL;
   retval->only_unallocated    = FALSE;
+  retval->is_nodisplay        = FALSE;
 
   if (parent != NULL)
     {
@@ -1324,7 +1343,8 @@ static GMenuTreeEntry *
 gmenu_tree_entry_new (GMenuTreeDirectory *parent,
 		      DesktopEntry       *desktop_entry,
 		      const char         *desktop_file_id,
-		      gboolean            is_excluded)
+		      gboolean            is_excluded,
+                      gboolean            is_nodisplay)
 {
   GMenuTreeEntry *retval;
 
@@ -1337,6 +1357,7 @@ gmenu_tree_entry_new (GMenuTreeDirectory *parent,
   retval->desktop_entry   = desktop_entry_ref (desktop_entry);
   retval->desktop_file_id = g_strdup (desktop_file_id);
   retval->is_excluded     = is_excluded != FALSE;
+  retval->is_nodisplay    = is_nodisplay != FALSE;
 
   return retval;
 }
@@ -2033,6 +2054,15 @@ add_filename_include (const char     *desktop_file_id,
     }
 }
 
+static void
+is_dot_directory (const char   *basename,
+		  DesktopEntry *entry,
+		  gboolean     *has_dot_directory)
+{
+  if (!strcmp (basename, ".directory"))
+    *has_dot_directory = TRUE;
+}
+
 static gboolean
 add_menu_for_legacy_dir (MenuLayoutNode *parent,
                          const char     *legacy_dir,
@@ -2042,21 +2072,31 @@ add_menu_for_legacy_dir (MenuLayoutNode *parent,
 {
   EntryDirectory  *ed;
   DesktopEntrySet *desktop_entries;
+  DesktopEntrySet *directory_entries;
   GSList          *subdirs;
   gboolean         menu_added;
+  gboolean         has_dot_directory;
 
   ed = entry_directory_new_legacy (DESKTOP_ENTRY_INVALID, legacy_dir, legacy_prefix);
   if (!ed)
     return FALSE;
 
   subdirs = NULL;
-  desktop_entries = desktop_entry_set_new ();
+  desktop_entries   = desktop_entry_set_new ();
+  directory_entries = desktop_entry_set_new ();
+
   entry_directory_get_flat_contents (ed,
                                      desktop_entries,
-                                     NULL,
+                                     directory_entries,
                                      &subdirs);
   entry_directory_unref (ed);
 
+  has_dot_directory = FALSE;
+  desktop_entry_set_foreach (directory_entries,
+			     (DesktopEntrySetForeachFunc) is_dot_directory,
+			     &has_dot_directory);
+  desktop_entry_set_unref (directory_entries);
+
   menu_added = FALSE;
   if (desktop_entry_set_get_count (desktop_entries) > 0 || subdirs)
     {
@@ -2080,21 +2120,24 @@ add_menu_for_legacy_dir (MenuLayoutNode *parent,
       menu_layout_node_append_child (menu, node);
       menu_layout_node_unref (node);
 
-      node = menu_layout_node_new (MENU_LAYOUT_NODE_DIRECTORY);
-      if (relative_path != NULL)
-        {
-          char *directory_entry_path;
+      if (has_dot_directory)
+	{
+	  node = menu_layout_node_new (MENU_LAYOUT_NODE_DIRECTORY);
+	  if (relative_path != NULL)
+	    {
+	      char *directory_entry_path;
 
-          directory_entry_path = g_strdup_printf ("%s/.directory", relative_path);
-          menu_layout_node_set_content (node, directory_entry_path);
-          g_free (directory_entry_path);
-        }
-      else
-        {
-          menu_layout_node_set_content (node, ".directory");
-        }
-      menu_layout_node_append_child (menu, node);
-      menu_layout_node_unref (node);
+	      directory_entry_path = g_strdup_printf ("%s/.directory", relative_path);
+	      menu_layout_node_set_content (node, directory_entry_path);
+	      g_free (directory_entry_path);
+	    }
+	  else
+	    {
+	      menu_layout_node_set_content (node, ".directory");
+	    }
+	  menu_layout_node_append_child (menu, node);
+	  menu_layout_node_unref (node);
+	}
 
       if (desktop_entry_set_get_count (desktop_entries) > 0)
 	{
@@ -2621,7 +2664,6 @@ gmenu_tree_execute_moves (GMenuTree      *tree,
   MenuLayoutNode *child;
   gboolean        need_remove_dups;
   GSList         *move_nodes;
-  GSList         *prev;
   GSList         *tmp;
 
   need_remove_dups = FALSE;
@@ -2651,7 +2693,6 @@ gmenu_tree_execute_moves (GMenuTree      *tree,
       child = menu_layout_node_get_next (child);
     }
 
-  prev = NULL;
   tmp = move_nodes;
   while (tmp != NULL)
     {
@@ -2986,9 +3027,10 @@ entries_listify_foreach (const char         *desktop_file_id,
   directory->entries =
     g_slist_prepend (directory->entries,
 		     gmenu_tree_entry_new (directory,
-					  desktop_entry,
-					  desktop_file_id,
-					  FALSE));
+                                           desktop_entry,
+                                           desktop_file_id,
+                                           FALSE,
+                                           desktop_entry_get_no_display (desktop_entry)));
 }
 
 static void
@@ -3001,7 +3043,8 @@ excluded_entries_listify_foreach (const char         *desktop_file_id,
 		     gmenu_tree_entry_new (directory,
 					   desktop_entry,
 					   desktop_file_id,
-					   TRUE));
+					   TRUE,
+                                           desktop_entry_get_no_display (desktop_entry)));
 }
 
 static GMenuTreeDirectory *
@@ -3161,8 +3204,9 @@ process_layout (GMenuTree          *tree,
                   }
               }
 
-            menu_verbose ("Processed <Directory> new directory entry = %p\n",
-                          directory->directory_entry);
+            menu_verbose ("Processed <Directory> new directory entry = %p (%s)\n",
+                          directory->directory_entry,
+			  directory->directory_entry? desktop_entry_get_path (directory->directory_entry) : "null");
           }
           break;
 
@@ -3218,9 +3262,14 @@ process_layout (GMenuTree          *tree,
     {
       if (desktop_entry_get_no_display (directory->directory_entry))
         {
-          menu_verbose ("Not showing menu %s because NoDisplay=true\n",
+          directory->is_nodisplay = TRUE;
+
+          if (!(tree->flags & GMENU_TREE_FLAGS_INCLUDE_NODISPLAY))
+            {
+              menu_verbose ("Not showing menu %s because NoDisplay=true\n",
                         desktop_entry_get_name (directory->directory_entry));
-          deleted = TRUE;
+              deleted = TRUE;
+            }
         }
 
       if (!desktop_entry_get_show_in_gnome (directory->directory_entry))
@@ -3267,7 +3316,8 @@ process_layout (GMenuTree          *tree,
           delete = TRUE;
         }
 
-      if (desktop_entry_get_no_display (entry->desktop_entry))
+      if (!(tree->flags & GMENU_TREE_FLAGS_INCLUDE_NODISPLAY) &&
+          desktop_entry_get_no_display (entry->desktop_entry))
         {
           menu_verbose ("Deleting %s because NoDisplay=true\n",
                         desktop_entry_get_name (entry->desktop_entry));
@@ -3477,7 +3527,7 @@ merge_entry (GMenuTree          *tree,
 static void
 merge_entry_by_id (GMenuTree          *tree,
 		   GMenuTreeDirectory *directory,
-		   const char        *file_id)
+		   const char         *file_id)
 {
   GSList *tmp;
 
@@ -3501,61 +3551,104 @@ merge_entry_by_id (GMenuTree          *tree,
     }
 }
 
+static inline gboolean
+find_name_in_list (const char *name,
+		   GSList     *list)
+{
+  while (list != NULL)
+    {
+      if (!strcmp (name, list->data))
+	return TRUE;
+
+      list = list->next;
+    }
+
+  return FALSE;
+}
+
 static void
 merge_subdirs (GMenuTree          *tree,
-	       GMenuTreeDirectory *directory)
+	       GMenuTreeDirectory *directory,
+	       GSList             *except)
 {
+  GSList *subdirs;
   GSList *tmp;
 
   menu_verbose ("Merging subdirs in directory '%s'\n", directory->name);
 
-  directory->subdirs = g_slist_sort (directory->subdirs,
-				     (GCompareFunc) gmenu_tree_directory_compare);
+  subdirs = directory->subdirs;
+  directory->subdirs = NULL;
 
-  tmp = directory->subdirs;
+  subdirs = g_slist_sort (subdirs,
+			  (GCompareFunc) gmenu_tree_directory_compare);
+
+  tmp = subdirs;
   while (tmp != NULL)
     {
       GMenuTreeDirectory *subdir = tmp->data;
 
-      merge_subdir (tree, directory, subdir, &directory->default_layout_values);
-      gmenu_tree_item_unref (subdir);
+      if (!find_name_in_list (subdir->name, except))
+	{
+	  merge_subdir (tree, directory, subdir, &directory->default_layout_values);
+	  gmenu_tree_item_unref (subdir);
+	}
+      else
+	{
+	  menu_verbose ("Not merging directory '%s' yet\n", subdir->name);
+	  directory->subdirs = g_slist_append (directory->subdirs, subdir);
+	}
 
       tmp = tmp->next;
     }
 
-  g_slist_free (directory->subdirs);
-  directory->subdirs = NULL;
+  g_slist_free (subdirs);
+  g_slist_free (except);
 }
 
 static void
 merge_entries (GMenuTree          *tree,
-	       GMenuTreeDirectory *directory)
+	       GMenuTreeDirectory *directory,
+	       GSList             *except)
 {
+  GSList *entries;
   GSList *tmp;
 
   menu_verbose ("Merging entries in directory '%s'\n", directory->name);
 
-  directory->entries = g_slist_sort (directory->entries,
-				     (GCompareFunc) gmenu_tree_entry_compare);
+  entries = directory->entries;
+  directory->entries = NULL;
 
-  tmp = directory->entries;
+  entries = g_slist_sort (entries,
+			  (GCompareFunc) gmenu_tree_entry_compare);
+
+  tmp = entries;
   while (tmp != NULL)
     {
       GMenuTreeEntry *entry = tmp->data;
 
-      merge_entry (tree, directory, entry);
-      gmenu_tree_item_unref (entry);
+      if (!find_name_in_list (entry->desktop_file_id, except))
+	{
+	  merge_entry (tree, directory, entry);
+	  gmenu_tree_item_unref (entry);
+	}
+      else
+	{
+	  menu_verbose ("Not merging entry '%s' yet\n", entry->desktop_file_id);
+	  directory->entries = g_slist_append (directory->entries, entry);
+	}
 
       tmp = tmp->next;
     }
 
-  g_slist_free (directory->entries);
-  directory->entries = NULL;
+  g_slist_free (entries);
+  g_slist_free (except);
 }
 
 static void
 merge_subdirs_and_entries (GMenuTree          *tree,
-			   GMenuTreeDirectory *directory)
+			   GMenuTreeDirectory *directory,
+			   GSList             *except_subdirs,
+			   GSList             *except_entries)
 {
   GSList *items;
   GSList *tmp;
@@ -3578,22 +3671,42 @@ merge_subdirs_and_entries (GMenuTree          *tree,
 
       if (gmenu_tree_item_get_type (item) == GMENU_TREE_ITEM_DIRECTORY)
 	{
-	  merge_subdir (tree,
-			directory,
-			GMENU_TREE_DIRECTORY (item),
-			&directory->default_layout_values);
+	  if (!find_name_in_list (GMENU_TREE_DIRECTORY (item)->name, except_subdirs))
+	    {
+	      merge_subdir (tree,
+			    directory,
+			    GMENU_TREE_DIRECTORY (item),
+			    &directory->default_layout_values);
+	      gmenu_tree_item_unref (item);
+	    }
+	  else
+	    {
+	      menu_verbose ("Not merging directory '%s' yet\n",
+			    GMENU_TREE_DIRECTORY (item)->name);
+	      directory->subdirs = g_slist_append (directory->subdirs, item);
+	    }
 	}
       else
 	{
-	  merge_entry (tree, directory, GMENU_TREE_ENTRY (item));
+	  if (!find_name_in_list (GMENU_TREE_ENTRY (item)->desktop_file_id, except_entries))
+	    {
+	      merge_entry (tree, directory, GMENU_TREE_ENTRY (item));
+	      gmenu_tree_item_unref (item);
+	    }
+	  else
+	    {
+	      menu_verbose ("Not merging entry '%s' yet\n",
+			    GMENU_TREE_ENTRY (item)->desktop_file_id);
+	      directory->entries = g_slist_append (directory->entries, item);
+	    }
 	}
 
-      gmenu_tree_item_unref (item);
-
       tmp = tmp->next;
     }
 
   g_slist_free (items);
+  g_slist_free (except_subdirs);
+  g_slist_free (except_entries);
 }
 
 static void
@@ -3643,6 +3756,56 @@ get_layout_info (GMenuTreeDirectory *directory)
   return NULL;
 }
 
+static GSList *
+get_subdirs_from_layout_info (GSList *layout_info)
+{
+  GSList *subdirs;
+  GSList *tmp;
+
+  subdirs = NULL;
+
+  tmp = layout_info;
+  while (tmp != NULL)
+    {
+      MenuLayoutNode *node = tmp->data;
+
+      if (menu_layout_node_get_type (node) == MENU_LAYOUT_NODE_MENUNAME)
+	{
+	  subdirs = g_slist_append (subdirs,
+				    (char *) menu_layout_node_get_content (node));
+	}
+
+      tmp = tmp->next;
+    }
+
+  return subdirs;
+}
+
+static GSList *
+get_entries_from_layout_info (GSList *layout_info)
+{
+  GSList *entries;
+  GSList *tmp;
+
+  entries = NULL;
+
+  tmp = layout_info;
+  while (tmp != NULL)
+    {
+      MenuLayoutNode *node = tmp->data;
+
+      if (menu_layout_node_get_type (node) == MENU_LAYOUT_NODE_FILENAME)
+	{
+	  entries = g_slist_append (entries,
+				    (char *) menu_layout_node_get_content (node));
+	}
+
+      tmp = tmp->next;
+    }
+
+  return entries;
+}
+
 static void
 process_layout_info (GMenuTree          *tree,
 		     GMenuTreeDirectory *directory)
@@ -3659,8 +3822,8 @@ process_layout_info (GMenuTree          *tree,
 
   if ((layout_info = get_layout_info (directory)) == NULL)
     {
-      merge_subdirs (tree, directory);
-      merge_entries (tree, directory);
+      merge_subdirs (tree, directory, NULL);
+      merge_entries (tree, directory, NULL);
     }
   else
     {
@@ -3707,15 +3870,22 @@ process_layout_info (GMenuTree          *tree,
 		  break;
 
 		case MENU_LAYOUT_MERGE_MENUS:
-		  merge_subdirs (tree, directory);
+		  merge_subdirs (tree,
+				 directory,
+				 get_subdirs_from_layout_info (tmp->next));
 		  break;
 
 		case MENU_LAYOUT_MERGE_FILES:
-		  merge_entries (tree, directory);
+		  merge_entries (tree,
+				 directory,
+				 get_entries_from_layout_info (tmp->next));
 		  break;
 
 		case MENU_LAYOUT_MERGE_ALL:
-		  merge_subdirs_and_entries (tree, directory);
+		  merge_subdirs_and_entries (tree,
+					     directory,
+					     get_subdirs_from_layout_info (tmp->next),
+					     get_entries_from_layout_info (tmp->next));
 		  break;
 
 		default:
